\ 3iForth.fs -- 3-Instruction Forth for AVR Atmel 328 / Arduino Nano on GForth 0.7.0 (Windows PC)
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22  FOR VERSION HISTORY see 3iForth_history.txt file
\ Article: http://www.mathscitech.org/articles/3iForth
\ c. 150 lines of Forth code (excluding constants and comments) - 2021-08-10---Tue---02:53

\ Hardware (target) - Arduino Nano (Atmega328P uc with USB mini)
\ Firmware: 3iForth.avr compiled & uploaded using Arduino IDE 1.8.9 & "Old Bootloader"
\ Host (PC): 3iForth.fs running in GForth 0.7.0 on Windows PC

\ Description: On startup/rest, Nano sets up serial port (9600 baud), initializes data stack, and waits for instructions over serial from PC via USB mini cable

\ USAGE:
\   1. Connect PC to Nano using USB mini cable
\   2. On PC: Run RealTerm with command line setup: baud=9600 port=1  half duplex  This sets up the serial port correctly 
\   3.        Run Gforth: include 3iForth.fs


include 3iCore.fs


\ OPENING SERIAL CONNECTION TO NANO - open RealTerm first  (16 lines)
\      TODO: need to add GForth code to configure COM port correctly to remove dependency on RealTerm
variable COMhandle  \ Windows handle for COM port provided by create-file command
: COM@ ( -- handle ) COMhandle @ ;
create  comstringdata 'c' c, 'o' c, 'm' c, '0' c,            \ COM port string
: attach-port ( port# -- ) '0' +   comstringdata 3 + c!  ;
: comstring ( -- c-adr nlen ) comstringdata 4 ;   \ ex: comstring type
: com-notify ( handle errcode -- )  \ responds to create-file with success notification or throwing error
    dup 0 = if drop COMhandle !  comstring type ."  opened - " 
            else cr ." **Error: check device is plugged into " comstring type cr throw
            then ;
: [OPEN-COM] ( port# --  )         \ Opens COM port in binary mode
    attach-port comstring r/w bin create-file com-notify ;
: close-com ( -- )  COM@ close-file throw comstring type ."  closed." cr ;
: reload ( -- )   \ ensure com port is closed before reloading which clobbers COM@ so need to open-com port again
  COM@ 0> if close-com then
  s" 3iForth.fs" included ;


\ READING FROM SERIAL COMM RECEIVE BUFFER (8 lines)
\    Design: serial messages from Nano are held in byte receive buffer on PC for user program to read
\    Design: 80ms delay required between sending serial command and receiving serial response from uC. 60ms intermittent, 25 !works
create RX-BUFFER 255 bytes allot
: read-com ( n -- cnt )   \ read N bytes from COM port into rx-buffer,  ex: 50 read-com show-rx
    RX-BUFFER swap COM@ read-file throw ;
: .combytes ( cnt -- cnt ) dup . ." bytes on port" cr ; 
: show-rx ( cnt -- cnt )  \ prints rcv buffer in decimal, ex: 255 show-rx,  ./hex. prints in decimal/hex
    dup 0 > if  cr ." rcv: "  RX-BUFFER over 0 do c@+  hex.  loop drop ( else .combytes ) then ;
: rxN ( N -- cnt ) 80 ms read-com ( .combytes ) show-rx ;     
: rx  ( -- ) 50 rxN drop ;

\   Design: built-in open-com for easy initialization - opens and responds
: open-com ( port# -- ) [open-com] 2000 ms  5 rxN drop ;  \ 1500ms delay misses response, 2000 works


\ TRANSMITTING DATA VIA SERIAL COMM TRANSMIT BUFFER  (5 lines)
\    Design: only tx transmits.  everything else prepares the TX-BUFFER (fewer comms -> more efficient)
\    Design: transfers are BYTE transfers.  Anything larger, e.g. multi-byte numbers, need to be broken down to tx 1 byte at a time.
create TX-BUFFER 20 bytes allot     \ CODE REVIEW: is 20 bytes enough?
: show-tx ( -- )  TX-BUFFER 20 0 cr ." TX-BUFFER>: " do c@+ . loop drop ;  \ change . to hex. to show hex values
\ CODE REVIEW: no timing delays needed?
: tx ( adr'  -- )   \ send over serial [len=adr'-adr] bytes of TX-BUFFER in FIFO left-to-right order 
    TX-BUFFER swap over -  COM@ write-file throw COM@ flush-file throw  ;  
: tx1 ( byte -- ) TX-BUFFER c!+ tx ;  \ sends 1 byte over serial


\ CORE 3-INSTRUCTION (5-INSTRUCTION) FORTH COMMANDS, distinct from exported chip instructions  (17 lines)
\    Design:  these 5 bytes must preface any data which follows.  The target chip will only respond to these instructions.
1 constant CMD_XSTORE     \ push TOS (local) onto remote stack, see nstore for transmit comms efficiency
3 constant CMD_XFETCH     \ pop top byte off remote stack and put onto local stack
4 constant CMD_XCALL      \ call function ID that must already be on the remote stack along with all parameters
\    Design:   two additional instructions for convenience/diagnostics
2 constant CMD_XDROP      \ pop top byte off remote stack and discards
0 constant CMD_X_S        \ maps to stack_len() in 3iForth.avr - displays remote stack; queries chip & reports stack size in bytes

\ XSTORE - builds instructions + parameters into TX-BUFFER ready for serial transmission in one burst w/ tx
\    Design: functions acting on TX-BUFFER accept adr and output adr' next address to allow chaining easily
: [storebyte]  ( byte  adr -- adr' )  CMD_XSTORE swap c!+ c!+ ;         \ prepares instruction: store 1 byte
\    Design: serial transmit is a QUEUE (works from start of TX-BUFFER treated as FIFO)
\    Design: nstore uses roll to duplicate local onto remote stack, which allows testing chip as if it is local without risking pins
\    Design: nstore is useful for interfaces where user types in instructions to right, and wants execution left to right
\    Design: we use nstore to maintain same stack diagram in host (.fs) as target (.avr) with one EXCEPTION
\    Design: EXCEPTION - for unrestricted byte transfers eg 3iI2C.avr cmdOLED and dataOLED words we use revstore
: [nstore]     ( w1 w2 ... wn adr n -- adr' )   \ prepare to store w1 ... wn on target -- MAX 10 elements!
    swap over 0 do over i - 1 + roll swap [storebyte] loop swap drop ;
: nstore       ( n -- )  TX-BUFFER swap [nstore]   tx ;  \ transmit to target
\    Design: revstore reverses order of stack on remote machine - used in e.g. I2C.fs  (recall: LIFO=stack, FIFO=queue)
: [revstore]   ( w1 w2 ... wn adr n -- adr' )  0 do [storebyte] loop ;  \ prepares to store wn ... w1 on target -- MAX 10 elements!
: revstore ( n -- )  TX-BUFFER swap [revstore] tx ;  \ transmit to target

: xstore  ( a -- )    1 nstore  ;                    \ store byte onto remote stack using TX-BUFFER
: xfetch  ( -- val )     CMD_XFETCH  tx1   80 ms   1 read-com drop   RX-BUFFER c@ ;
: xcall   ( instr-id -- ) xstore CMD_XCALL tx1 ;     \ call exported chip instruction (see .avr fcall jump table)

: x_s     ( -- )      cr CMD_X_S     tx1   rx   xfetch ." (" . ." )" ;  
: xdrop   ( -- )         CMD_XDROP   tx1 ;
: 2xdrop                 xdrop xdrop ;
        

\ EXPORTED MICROCONTROLLER FUNCTIONS INTERFACE
\   Design: the user thinks they live in the target and calls the uC exported functions directly using Forth
\   Design: these constants must match the jumptable entries 3iForth.avr which is burned onto the uC chip
4   constant CALL_SETREG
$E  constant CALL_CHECKREG
0   constant CALL_PINMODE
$B  constant CALL_PORT4MODE
1   constant CALL_PINWRITE
$C  constant CALL_PORT4WRITE
$A  constant CALL_PINPWM    \ write simulated analog out by varying duty cycle of digital pinwrite
$18 constant CALL_SETTIMER  \ changes PWM timer
3   constant CALL_TONE
7   constant CALL_PINREAD
$19 constant CALL_PINSIGNAL
$D  constant CALL_PORT4READ
$1A constant CALL_PORT4SIGNAL
$1E constant CALL_SHIFTINBYTE
9   constant CALL_PINREADLEVEL
\ Application specific capabilities
\  $1C constant CALL_SHIFTOUTBYTE
\  $1D constant CALL_SHIFTOUT32  \ see 3iShift.fs
5   constant CALL_SWITCHKEY  \ digital piano only (piano.avr and piano-4x4.avr)
6   constant CALL_BUTTONSRECORD     \ digital piano only (piano.avr and piano-4x4.avr)
\  8 call_keypad     \ poll keypad (in 3iKeypad.fs OR 3iDrumPad.fs)
\ $F through $16 are  8 instructions related to use of OLED displays (oled.avr)
\ $17 is motor movement instruction (motor.avr)
\ $1B pollpianokey - piano.fs
\ $1F through $21 are 3 instructions related to use of OLED displays with SSD1306 using only I2C, i.e. without driver software
\ Obsolete - remove
\  2   constant CALL_DELAY   -- NOT USED - in host Forth delays using ms.  In target, timing specific, is written into the commands



\ GFORTH CODE TO EASILY CONTROL EXPORTED MICROCONTROLLER FUNCTIONS (100 lines, including comments)
\   Design: application specific problems are solved in various .fs (Forth solution) files creating simplified yet powerful customer dictionaries, Chuck Moore's PROBLEM ORIENTED LANGUAGE, on the host.

\   Design: Registers store configuration parameters interactively which application specific commands use, e.g. piano plays default duration, see 3iForth.avr for list
: setreg ( val16 reg# -- )   \ ex: 440 0 setreg    sets register0 to 16-bit value   ( e.g. musical note A4=440 Hz )
    swap lohi rot   3 nstore CALL_SETREG xcall ;
: checkreg ( reg# --  ) xstore CALL_CHECKREG xcall    rx ;  \ serial shows contents: ( -- rx>: hi lo )


\ ARDUINO NANO PINS - pins 1-13 are digital.  A0-A5 (14-19) are analog/digital.  A6-A7 (20-21) are AnalogRead only.
0 constant INPUT    \ no built-in Arduino constant for INPUT_PULLDOWN - CODE REVIEW: check that input is equivalent to pulldown
1 constant OUTPUT
2 constant INPUT_PULLUP  \ e.g. for active low buttons
: pinMode ( mode pin -- )   \ set pin into mode, ex: OUTPUT 9 pinmode
    2 nstore CALL_PINMODE xcall 60 ms ;   \ settle time on pin req'd when changing mode.  60 ms works
: out    ( pin -- )       OUTPUT swap pinmode ;  \ ex: 9 out
: in     ( pin -- )        INPUT swap pinmode ;
: pullup ( pin -- ) INPUT_PULLUP swap pinmode ;  \ ex: 2 pullup

1 constant HIGH_SIGNAL
0 constant LOW_SIGNAL
: pinwrite ( level pin -- )  2 nstore  CALL_PINWRITE xcall ;  \ ex: 1 9 pinwrite   0 9 pinwrite
: high ( pin -- )  HIGH_SIGNAL swap pinwrite ;
: low ( pin -- )   LOW_SIGNAL swap pinwrite ;
: lo low ;
: hi high ;
\ for instantaneous signals  (used e.g. in 3iShift.fs)
: pulse-hi ( pin -- ) dup hi 1 ms lo ;   \ rising edge pulse (for clk and latch)
: pulse-lo ( pin -- ) dup lo 1 ms hi ;   \ falling edge pulse (for clear)

: pwm ( val8 pin -- ) 2 nstore CALL_PINPWM xcall ;   \ 8-bit value (0-255)  pins 3,9,10,11 (490 Hz), pins 5,6 (980 Hz)
: testpwm ( -- ) 50 10 pwm 500 ms  255 10 pwm 500 ms 0 10 pwm 500 ms ;
\ max is -212, max before hitting stop is -208, min is -2, 0 is weird
: settimer ( prescale timer -- ) 2 nstore CALL_SETTIMER xcall ;  \ Timer 0 pins 5,6,  Timer 1 pins 9,10, Timer 2 pins 3, 11

\   Design: Registers 1-13 reserved in 3iForth.avr to hold frequencies for a 13-key keyboard, allowing programmable change of scale
\   Design: Registers 14, 15 reserved in 3iForth.avr to hold default duration and rest following piano keyboard button press
: [tone] ( freq-hi freq-lo dur-hi dur-lo pin -- ) 5 nstore CALL_TONE xcall ;
   \ ex: OUTPUT 13  (if speaker connected to pin 13)
   \ ex: 440 lohi 500 lohi _speaker [tone]
   \ https://www.arduino.cc/reference/en/language/functions/advanced-io/tone/
   \ https://github.com/bhagman/Tone#ugly-details

: pinread ( pin -- 0/1 )   \ digitalRead() - with realtime delay (COMPLICATION)  -- WARNING: set pin mode to INPUT (active high) or PULLUP (active low)
	xstore CALL_PINREAD xcall
	120 ms   \ 50 ms seems to work ( 0 failures out of 30 trials ), if processor is also writing to display, then 120ms is enough
	1 read-com if RX-BUFFER c@ else cr ." FAIL pinread: no bytes" cr then ;
: poll-pin ( pin -- ) begin dup pinread . key? until ." Done!" drop key drop ;   \ for active low.  if active high remove the not
: pinsignal ( pin -- ) \ read pin until get a signal (0 or 1)
    xstore CALL_PINSIGNAL xcall
    begin 100 ms 1 read-com until RX-BUFFER c@ ;  \ BLOCKING READ

: pinreadLevel ( Apin -- counts )   \ analogRead() -- NOTE: use ANALOG pin (A0-A7, 14-21) - result in 10 bit counts (0 to 1024)
    \ WARNING: set pin mode to INPUT
    \ Ref: https://www.arduino.cc/reference/en/language/functions/analog-io/analogread/
	xstore CALL_PINREADLEVEL xcall
	60 ms   \ 50 ms seems to work ( 0 failures out of 30 trials )
	2 read-com if RX-BUFFER c@+ 8 lshift swap c@ + else cr ." FAIL pinread: no bytes" cr then ;
: Vref5 ( cnts -- mV ) 49 10 */ ;  \ 4.9mV per count  (5V=1024 counts)
: poll-level ( Apin -- ) begin dup pinreadLevel Vref5 . cr 100 ms key? until ." Done!" drop ;
: plot-level ( Apin -- ) begin dup pinreadLevel Vref5 dup 100 / cr spaces ." XX  " . 100 ms key? until ." Done!" drop ;


\ 4-bit Port Control   
15 constant REG_PORTAOUT  \ reserved register values in 3iForth.avr
16 constant REG_PORTBIN
17 constant REG_PORTCOUT
18 constant REG_PORTDIN
\ Ex:  5 4 3 2  pins for output port, lsb@TOS.   Ex: 11 10 9 8 pins for input port, lsb@TOS
\   Design: !!WARNING!! - can only use pins up to 15 $F with join4
: join4 ( msb c b lsb -- mcbl-16bits )  \  ex: 5 4 3 2 -- $5432 (in hex) combines 4 bytes (pins) to 16-bit value for setreg
    swap 4 lshift or
    swap 8 lshift or
    swap $C lshift or ;
: [port4out] ( d c b a reg# -- )  \ ex: 5 4 3 2  port4out, lsb TOS      Test: reg_portout checkreg   WARNING: MAX PIN IS 15
    >r join4 r@ setreg 100 ms   OUTPUT r>  2 nstore CALL_PORT4MODE xcall ;
: [port4write] ( val reg# -- ) 2 nstore 2 nstore CALL_PORT4WRITE xcall ; \ writes hex digit to 4-bit port in reg#
: port4out  ( d c b a -- )  REG_PORTAOUT [port4out] ;
: port4Cout ( d c b a -- )  REG_PORTCOUT [port4out] ;
: port4write    ( hex -- )  REG_PORTAOUT [port4write] ;           
: port4Cwrite   ( hex -- )  REG_PORTCOUT [port4write] ;
: [port4in] ( d c b a reg# -- ) \ ex: 11 10 9 8  port4in, lsb TOS     Test: reg_portin checkreg
    >r join4 r@ setreg 80 ms  INPUT r> 2 nstore CALL_PORT4MODE xcall ;
: port4in  ( d c b a -- )  REG_PORTBIN  [port4in] ;
: port4Din  ( d c b a -- ) REG_PORTDIN  [port4in] ;
: [port4read] ( reg# -- hex ) \ reads hex digit from 4-bit port
    2 nstore CALL_PORT4READ  xcall   100 ms  \ 60 ms seems to work
    1 read-com if RX-BUFFER c@ else cr ." FAIL port4read: no bytes" cr then ;
: [port4signal] ( reg# -- ) \ polls input port until something is received (on serial)
    xstore CALL_PORT4SIGNAL xcall
    begin 100 ms 1 read-com until RX-BUFFER c@ ; \ BLOCKING READ
: port4read ( -- hex ) REG_PORTBIN [port4read] ;
: poll-port ( -- ) begin port4read not4 dup if . cr else ." ." drop then key? until ." Done!" ;  \ use not4 for active low
: port4signal ( -- ) REG_PORTBIN [port4signal] ; 

: buttonsrecord  ( on1/off0 -- )    \ turns on/off piano serial interface -- when a note is played, piano sends digital note data
    xstore CALL_BUTTONSRECORD xcall ; 


\ HELP FILE ( * HELP FILES * )
: help
    cr
    cr ." help "
    cr ."  ------------------------------------------------------------"
    cr ."     includes 3iCore.fs  for GForth core functionality for my Forths"
    cr ."  3-instruction Forth for Atmel 328 chip:"
    cr ."     STARTING/STOPPING"
    cr ."       1 open-com  -  opens communication to port 1 and flushes response (should be $FF $FE / 255 254)"
    cr ."       close-com - to end session"
    cr ."       reload - to reload 3iForth.fs file after instruction edits - WARNING use this to avoid GForth lock-ups"
    cr 
    cr ."     OUTPUT PINS"
	cr ."       2 out         - sets pin 2 into output mode"
    cr ."       2 high        - sends 5V to pin 2  (choose pins 2 through 12)"
    cr ."       2 low         - sends 0V to pin 2  (choose pins 2 through 12)"
	cr ."	 INPUT PINS"
    cr ."       2 pullup     - sets pin 2 into pullup high mode, i.e. pin is then active low"
	cr ."       2 pinread "
	cr ."    LOWEST LEVEL FUNCTIONALITY"
    cr ."       host:     x_s    1 xstore   xdrop  xfetch .   <instr> xcall   "
    cr ."       target functions:   "
    cr ."          ms  pinmode   pinwrite   pinread "
    cr ."          pinwriteLevel pinreadLevel  setreg  checkreg  "
    cr ."          port4mode  port4write  port4read  tone  "
    cr 
    cr ."    Add-ons after 3iForth.fs:"
    cr ."       1/ include   3iI2C.fs          (for using OLED SSD1306 display without Adafruit software drivers)"
    cr ."       2/ include   3iKeypad.fs       (for 4x4 matrix keypad)"
    cr ."       3/ include   3iLights.fs       (for 12 LED lights board)"
    cr ."       4/ include   3iSound-4x4.fs    (for board with speaker)"
    cr ."       5/ include   3iBeats.fs   3iBeats_cookbook.fs   3iBeats_menu.fs  3iDrumPad.fs (for board with speaker and rate pot)"
    cr ."       6/ include   3iShift.fs        (for board with shift in or shift out registers)"
    cr ."       7/ include   3iPianoLights.fs  (for board with 32-piano-leds)"
    cr ."       8/ include   3iSonglist.fs     (CAN INCLUDE ONLY THIS -- IT auto-includes 4 dependencies)"
    cr ."       9/ include   3iArpegiattor.fs  3iGinan.fs  3iRecord.fs (for musical generation)"
    cr ."       9/ include   3iSensor.fs   3iMotor.fs   3iRadio.fs  (for laser, PIR sensors, steppers, servos, RC WiFi ...)"
    cr ."       10/ include  3iPort-4x4.fs     (for board with 4-to-16 and 16-to-4 encoder chips)"
    cr ."       obsolete:    3iOLED.fs         (for using OLED display WITH 17kB Adafruit software drivers)"
    cr cr
;

\ ======================
help
