( Sound Generation for 3-instruction Forth on Atmega328P uC / Arduino Nano board )
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22, 2020-07-25 edited, 2021-05-16---Sun---01:18

( * WORDS FOR COMPUTERIZED MUSIC * )

\ Hardware Options:
\   Buzzer: 5V,  + on pin, - on ground, use tone 440 --- loud sound
\   8 OHM Speaker with 250R or greater resistor.  resistor between pin and speaker+, speaker - to ground --- small sound
\   8 OHM Speaker with 2N2222 transistor for amplification.  
\       Wiring: 5V to collector, uC audio pin to base, spkr+ to emitter, spkr- to GRD with 100R resistor and 10uF capacitor in parallel in between-- big sound

\ Design: Music is a time series of NOTES/TONES (frequencies), their DURATIONS (ms), and the RESTS (ms) between them.
\ Design: Simplest tone generation specifies frequency and uses fixed duration and fixed rest, but this is not expressive.
\ Design: Better is 3 arrays holding note frequencies, note durations, and rests.  Play reads all three to form the instructions.
\ Design: Tone frequency is selected from a frequency table of 96 semitones (0-95), spanning X octaves.
\           freq takes index and returns the frequency.  Tone takes the frequency and produces the sound.
\           33 tone   or 0 freq tone
\           note = freq tone
\           ,,  ( for simplified 1..8 on selected scale )

\ Design: Could have two speakers and voices..., for now just one
variable SPEAKERpin  
: setspeaker ( pin -- ) SPEAKERpin ! ;   \ ex: 13 setspeaker
: speaker ( -- pin ) SPEAKERpin @ ;

\ Design: Duration and rest are read from variables.
variable dur_ms 
: beat dur_ms @ ; 
: setbeat ( n -- ) dur_ms ! ;    
: R1 beat 3 / ms ; \ semiquaver rest (1/16 note)
: R2 beat 2 * ms ; \ crotchet rest (1/4 note)
120 setbeat \ baseline tone duration ms  (sixteenth note - semiquaver)

\ Goal: Music tuning is a fascinating subject.  The programmable digital piano allows the exploration of this subject
\   Just (ratio) tuning:  https://en.wikipedia.org/wiki/Just_intonation

\ Design: Musical Tones 96 semitones (half steps) spanning 8 octaves (=12*8)
\ The mathematics of western music is multiplicative (equitempered scale popularized by Bach in the 1700s)
\ The musical scale has 12 semitones with each octave being double the frequency of the previous
\ By convention, there are 2 flats (b%, e%) and three sharps (c#, f#, g#)
\ Ratio between successive musical semitones is fixed as 12th-root of 2 = 2^(1/12) = 1.05946309436 (approx 6% higher freq than last)
\ Design: fixed freq table 384 bytes (=96*4) avoids cumulative tuning error vs fixed point multipl (1%, 5%, 15%, 40% for 1,2,3,4 octaves)
\ Design: For Arduino Atmega328P - It is not possible to generate tones lower than 31Hz (B0)
create piano   \ each cell in lookup table is 4 bytes
        \       0            2             4      5             7             9            11      
        \             1             3                    6             8             10
        \       C     d%     D      e%     E      F      F#     G      a%     A      b%     B
\ ( -1 )      16 ,   17 ,   18 ,   19 ,   21 ,   22 ,   23 ,   25 ,   26 ,   28 ,   29 ,   31 ,   \ 108 tones, except this too low
   ( 0 )      33 ,   35 ,   37 ,   39 ,   41 ,   44 ,   46 ,   49 ,   52 ,   55 ,   58 ,   62 ,  \ bass guitar lower E___
   ( 1 )      65 ,   69 ,   73 ,   78 ,   82 ,   87 ,   93 ,   98 ,  104 ,  110 ,  117 ,  124 ,  \ 6-string guitar low E__
   ( 2 )     131 ,  139 ,  147 ,  156 ,  165 ,  175 ,  185 ,  196 ,  208 ,  220 ,  233 ,  247 ,
   ( 3 )     262 ,  277 ,  294 ,  311 ,  330 ,  349 ,  370 ,  392 ,  415 ,  440 ,  466 ,  494 ,  \ middle C_
   ( 4 )     523 ,  554 ,  587 ,  622 ,  659 ,  698 ,  740 ,  784 ,  831 ,  880 ,  932 ,  988 , 
   ( 5 )    1047 , 1109 , 1175 , 1245 , 1319 , 1397 , 1480 , 1568 , 1661 , 1760 , 1865 , 1976 , 
   ( 6 )    2093 , 2217 , 2349 , 2489 , 2637 , 2794 , 2960 , 3126 , 3322 , 3520 , 3729 , 3951 , 
   ( 7 )    4186 , 4435 , 4699 , 4978 , 5274 , 5588 , 5920 , 6272 , 6645 , 7040 , 7459 , 7908 ,
\ Design: Chinese scales (pentatonic) use different frequences from Western frequencies
\ Design: Frequencies for Chinese scales are written directly into the frequency table
\ Design: Root note pointers
   ( 8 yu )    440 , 529  , 594  , 660  , 793  , 440 2 * , 529 2 * , 594 2 *   ,      \ Yu scale on Huang Zhong base freq
\               A    C'+6   D'+7   E'+1   G'+9   A'        C''+11    D''+13
   ( 9 shang ) 495 , 557  , 669  , 743  , 892  , 495 2 * , 557 2 * , 669 2 *   ,         \ Shang scale on Da Cu  base freq ( Snoopy theme)
\              B+1   c#'+3  e'+10  f#'+3  a'+12  b''+2     c#''+5    e''+19
  ( 10 jue )   705 , 847  , 952  , 1129 , 1270 , 705 2 * , 847 2 * , 952 2 *   , \ Jue scale on Yi Ze base freq 
  ( 11 zhi )   793 , 892  , 1071 , 1189 , 1338 , 793 2 * , 892 2 * , 1071 2 *  ,       \ Zhi scale on Wu Yi base freq

\ index into frequency table for Chinese scales
96  constant yu-root
104 constant shang-root
112 constant jue-root
120 constant zhi-root
\ 128


\ Design: keeping sound and lights separate in fundamental words.  See joint interface in 3iSonglist.fs
\ Design: interface for 1..8 notation is in 3iSonglist.fs

: freq ( note-index -- ) cells piano + @ ;  \ ex: 36 freq tone = C_ (middle C)
: f2i  ( freq -- index )  
   >r piano begin @+ r@ = until      \ search array for freq 
   piano -  4 bytes / 1 -  r> drop ; \ calculate index
: tone ( freq -- ) lohi beat lohi speaker [tone] R2 ;  \  ex: 440 tone  or 45 freq tone
: note ( note-index -- ) freq tone ;


\ =============== READING KEYBOARD PRESSES from serial
\ Design: to respond to keyboard presses need uC to send serial notifications when a key is pressed, turn on with 1 buttonsrecord (from 3iForth.fs)
\ Design: key press notifications are 6-byte tuples  (from 3iRecord.fs)
: fetch-note-data  ( <serial> -- numbytes ) 6 read-com ;  \ reads from serial buffer into rx-buffer, indicates if data is present (for polling use)
: show-note-data   ( -- ) rx-buffer 6 show-rx 2drop ;
: [parse-note] ( adr -- adr' val16 )  \ reads byte buffer parsed as little-endian words with low byte first to val16
    c@+ swap      c@+ 8 lshift     rot    or ;
: extract-note ( -- note-id ) rx-buffer 2 wwords + [parse-note] nip ;
: poll-note-data   ( <serial> -- note# ) begin fetch-note-data 0 > if  extract-note exitloop else continueloop then until ;
: >// 200 ms ;


\ ==================  NOTES

\ Design: # and % decorates denote sharp/flat and appear after note name
\ Design: _ and ' decorations denote lower and upper octaves on given note.  Middle C octave is _
\ Usage: c_ freq tone  or c_ note
0   constant  C____   \ octave lower 3
1   constant  c#____
1   constant  d%____ 
2   constant  D____ 
3   constant  d#____
3   constant  e%____
4   constant  E____ 
5   constant  F____ 
6   constant  f#____
6   constant  g%____
7   constant  G____      
8   constant  g#____
8   constant  a%____
9   constant  A____ 
10  constant  a#____
10  constant  b%____
11  constant  B____ 
12  constant  C___   \ octave lower 2
13  constant  c#___
13  constant  d%___ 
14  constant  D___ 
15  constant  d#___
15  constant  e%___
16  constant  E___ 
17  constant  F___ 
18  constant  f#___
18  constant  g%___
19  constant  G___      
20  constant  g#___
20  constant  a%___
21  constant  A___ 
22  constant  a#___
22  constant  b%___
23  constant  B___ 
24  constant  C__    \ octave lower 1
25  constant  c#__ 
25  constant  d%__ 
26  constant  D__  
27  constant  d#__ 
27  constant  e%__ 
28  constant  E__  
29  constant  F__    \  START of 32-key keyboard from F below middle C
30  constant  f#__ 
30  constant  g%__ 
31  constant  G__       
32  constant  g#__ 
32  constant  a%__  
33  constant  A__         
34  constant  a#__ 
34  constant  b%__ 
35  constant  B__  
36  constant  C_    \ middle octave -- Mddle C
37  constant  c#_  
37  constant  d%_  
38  constant  D_    
39  constant  d#_  
39  constant  e%_  
40  constant  E_    
41  constant  F_    
42  constant  f#_  
42  constant  g%_   
43  constant  G_   
44  constant  g#_      
44  constant  a%_  
45  constant  A_   
46  constant  a#_  
46  constant  b%_   
47  constant  B_   
48  constant  C'   \ octave upper 1
49  constant  c#'  
49  constant  d%'  
50  constant  D'   
51  constant  d#'  
51  constant  e%'  
52  constant  E'   
53  constant  F'   
54  constant  f#'  
54  constant  g%'  
55  constant  G'   
56  constant  g#'  
56  constant  a%'  
57  constant  A'   
58  constant  a#'  
58  constant  b%'  
59  constant  B'   
60  constant  C''    \ octave upper 2  --- END Of 32 key keyboards
61  constant  c#'' 
61  constant  d%'' 
62  constant  D''  
63  constant  d#'' 
63  constant  e%'' 
64  constant  E''  
65  constant  F''  
66  constant  f#'' 
66  constant  g%'' 
67  constant  G''  
68  constant  g#'' 
68  constant  a%'' 
69  constant  A''  
70  constant  a#'' 
70  constant  b%'' 
71  constant  B''  
72  constant  C'''   \ octave upper 3
73  constant  c#'''
73  constant  d%'''
74  constant  D''' 
75  constant  d#'''
75  constant  e%''' 
76  constant  E''' 
77  constant  F''' 
78  constant  f#'''
78  constant  g%'''
79  constant  G''' 
80  constant  g#'''
80  constant  a%'''
81  constant  A''' 
82  constant  a#'''
82  constant  b%'''
83  constant  B'''   \ 12-bits of frequency goes up to B'''
84  constant  C''''  \ octave upper 4

\ =================  SOUND DEMO ROUTINES and Effects ====================================
96 constant semitones  \ spanning 8 octaves
: upward  semitones 0 do i note loop ;
: downward semitones 0 do semitones 1 - i - note loop ;
: updown upward downward ;
: updown-fast 60 setbeat updown 120 setbeat ;
: updown-sfast 30 setbeat updown 120 setbeat ;
: updown-slide 15 setbeat updown 120 setbeat ;
: ring ( -- ) 440 lohi 500 lohi speaker [tone] ; 
: ring-ring  ( -- )  ring 600 ms ring ; \ ring tone


\ =================  APPLICATIONS =====================
\ 3iArpegiattor.fs : relies on board returning which key was pressed (turn on record/respond feature)
\ 3iPianoLights.fs : for 32 light board arranged in piano key layout (white/black keys) starting at F__ as 1 and up to C''

\ Piano PCB lights
\  create piano-led-map  \ maps led codes to led numbers, 0-based, to match switch/button numbering
\      $c c, $b c, $a c, $9 c, $8 c, $7 c, $6 c, $5 c, 
\      $4 c, $3 c, $2 c, $1 c, $0 c, $f c, $e c, $d c, 
\  : piano-test  \ tests each light, prompted by key presses
\      piano-led-map 16 0 do c@+ dup . _light_ key drop loop drop light| ;
\  : piano-flash
\      piano-led-map 16 0 do c@+ _light_ loop drop light| 50 ms ;
\  : piano-flash-n
\      0 do piano-flash loop ;
\  
\  : __ ( led# -- ) \ turn on appropriate light, 0-based
\      piano-led-map + c@ _light_ ;
\  : f2light ( freq -- light ) 
\      f2i 12 mod __ ;


\ Music theory:
\ Key signature: https://en.wikipedia.org/wiki/Key_signature
\   There are fifteen homogeneous key signatures ranging from 7 flats to 7 naturals to 7 sharps, with each successive flat or sharp placed on the note a perfect fifth below or above the previous one
\ Keys Major and Minor: https://en.wikipedia.org/wiki/Major_and_minor#Keys
\ In Western music, there are 12 semitones (half steps).  A scale picks out 8 of these.
\ A major scale has a HALFSTEP between notes 3 & 4 and between notes 7 & 8  (eg C major between E-F and B-C)
\ A minor scale puts the HALFSTEP between notes 2 & 3 and between notes 5 & 6 (eg D minor between E-F and A-B%)
\ Octave is a 2:1 ratio, Fifth is a 3:2 ratio
\ Intonation: 12-tone equal temperament is the most common tuning system in the West


\ * UNIVERSAL 8-KEY INSTRUMENT - NUMBERS TO NOTES *
\ Design: diatonic scale is an octet (8-tuple) of notes (piano sequence index) that sound good together
\ Design: using scales allows simplifying tunes using 1..8 references instead of the 13-key chromatic 
\ Design: Allows an 8 key piano to be able to play complex tunes by choosing the scale (key signature) and using note numbers 1-8
\ Design: Adding transposition by semitones, and movement of octaves, it becomes essentially a universal instrument
\ Setting up Key Signatures / Scales, and playing by reference against a scale
variable scale-ptr
: scale      ( -- a* )  scale-ptr @ ;
: set-scale  ( a* -- )  scale-ptr ! ;
: [,,]       ( 1-8# or 1-14# -- note )  1 - scale + c@ ;
:  ,,        ( 1-8# or 1-14# -- )       [,,] note ;        
: play-scale ( -- ) 9 1 do i ,, loop ;
\ Design: target for transpositions of scales
create tmp8 8 bytes allot


\ setup frequency-key mappings on target piano to enable changing scale/key signature/ethnic piano
create white-registers 0 c, 2 c, 4 c, 5 c, 7 c, 9 c, 11 c, 12 c,
create black-registers 1 c, 3 c, 6 c, 8 c, 10 c, 
: black-null  ( -- ) 5 0 do 0 black-registers i + c@ setreg loop ;
: xscale5 ( scale -- )  \ ex: cstd5 xscale5
    5 0 do dup i + c@ freq black-registers i + c@ setreg loop drop ;
: xscale8 ( scaleptr -- )  \ send scale to target  ex: cmaj8 xscale8
    8 0 do dup 
        i + c@ freq
        white-registers i + c@ 
        setreg 
    loop black-null drop ;

\ Design:  List 8 notes of key signature and write key using this
\ : test 9 1 do 8 i - roll . loop ;  \ 1 2 3 4 5 6 7 8   test
: write-key ( <create keyname> 1..8 -- ) 
    9 1 do 8 i - roll c, loop ;   \ writes key signature from supplied 8 notes
: set-key   ( keyptr -- ) dup set-scale xscale8 ;

\ Design: transposition by semitones
variable transp
: set-transp ( semitones -- ) transp c! ;
: get-transp ( semitones -- ) transp c@ ;
: ntranspose-list ( list n x -- )  \ transpose list by X (+ or -)  \ CODE REVIEW: will this work on songlist grammar? e.g. // /// >> << etc?
    set-transp dup 0 do dup i - roll get-transp + . loop drop ;
: transpose8  ( ptr x -- ) set-transp 8 0 do c@+ get-transp + swap loop drop ;

: >xscale8 ( 1..8 -- )  \ send stack scale to target  ex: cmaj8 -12 transpose8 >xscale8
   8 dup 0 do dup 
        i - roll freq
        white-registers i + c@ 
        setreg 
    loop black-null drop ;

: write-scale ( 1..8 -- ) tmp8 8 0 do 8 i - roll swap c!+ loop drop ;  \ takes diatonic list on stack and writes to tmp8 scale

\ Design: programmatically set keys to lower or higher octave
\ CODE REVIEW --- this doesn't set the other 5 keys
: ><< ( -- ) scale -12 transpose8 write-scale tmp8 set-key ;
: >>> ( -- ) scale  12 transpose8 write-scale tmp8 set-key ;


\ Design: For scales using non-western frequencies, e.g. Chinese scales, we write the new frequencies directly at the end of the frequency table
: write-freq-scale ( <create keyname> root-index ) 8 0 do dup c, 1 + loop drop ;

\ Design: shows the relative shifts that make up the scale
: show-scale  ( scaleptr -- ) 8 0 do c@+ 36 - . loop drop ;  \ ex: d%maj8 show-scale   cstd8 show-scale

\ WESTERN SCALES
\ Cstd8 is Bach's equitemper scale, do-re-me tuning
create cstd8     c_ d_ e_ f_ g_ a_ b_ c'  write-key    \ white keys
                 c#_ c, d#_ c, f#_ c, g#_ c, a#_ c, c#' c,  \ black keys - continuation to c#' to allow [,,,] to work
\ DESIGN: the 5 black keys are appended onto the 8 white keys
cstd8 8 + constant cstd5
: cstd  cstd8 set-key  cstd5 xscale5 ;

\ Cmaj8 starts on G  She'll be coming round the mountain when she comes...   O! Susanna (start on 3)
create cmaj8     G__ A__ C_ D_ E_ G_ A_ C'  write-key
: cmaj  cmaj8 set-key ; 

create dmin8     D_ E_ F_ G_ A_ b%_ c#' D'  write-key
: dmin  dmin8 set-key ;

create 2flats8 c_ d_ e%_ f_ g_ a_ b%_ c' write-key
: 2flats  2flats8 set-key ;
create irish8  c_ e%_ f_ g_ a_ b%_ c' d' write-key
: irish  irish8 set-key ;


create 3flats8 e%_ f_ g_ a%_ b%_ c' c#' e%' write-key
: 3flats 3flats8 set-key ;

\ 5flats is Db major - tuned on A -- five flats on D, E, G, A, B) with only C and F undecorated
          \ It is close to todi raga scale key but with F/F# and G%/G
          \ Repertoire:  I have a dream  and   (almost) Pir Nachi Ne garbi
create d%maj8   a%__  b%__ C_  d%_ e%_ F_ G%_ A%_   write-key
: d%maj   d%maj8 set-key ;
: 5flats  d%maj ;


\ EXOTIC SCALES

create phyrigian8 e_ f_ g_ a_ b_ c' d' e' write-key
: phyrigian  phyrigian8 set-key ;

\ INDIAN SCALES: Sa Re Ga Ma Pa Dha Ni Sa
\ Source: Special Topics in Music Theory, by Catherine Schmidt-Jones, Ch. 3.5 Indian Classical Music: Tuning and Ragas
\   https://cnx.org/contents/mviDomhA@5.66:bDLyHWWs@16/Indian-Classical-Music-Tuning-and-Ragas
\ A "that" is a scale of 7 notes -- why 7?  (7 celestial bodies - Sun, Moon, Mercury, Venus, Mars, Jupiter and Saturn)
\ Aside: 7 days a week: Babylonians of the 6th century... https://www.discovermagazine.com/planet-earth/why-are-there-seven-days-in-a-week
\ A raga chooses five, six, or all 7 notes of the that, and sticks to that, i.e. pentatonic raga, hexatonic raga, or full raga

\ FULL RAGAS
\ bhairav = full raga with two flats on D and A     NO EQUIVALENT WESTERN MAJOR SCALE
\ also arabic maqam - Hijaz Kar maqam - https://www.pianoscales.org/maqams.html
create     bhairav8   C_  d%_  E_  F_  G_  a%_  B_  C'  write-key   \ 1 2 5 6   8 9 12 13
: bhairav  bhairav8 set-key ;
: hijaz-kar bhairav ; \ arabic maqam
: byzantine bhairav ;

create     bhairavG8  G__ a%__ B__ C_  D_  e%_  f#_  G_  write-key  \ transposed 5 semitones down to G
: bhairavG bhairavG8 set-key ;

\ khamaj = raga with 1 flat on B                    EQUIVALENT TO F MAJOR
create   fmaj8  C_   D_   E_   F_   G_   A_   b%_   C'  write-key
: fmaj   fmaj8 set-key ;
: khamaj fmaj ;

\ kafi  = raga with two flats on B and E            EQUIVALENT TO B%-MAJ   ex: unchaare kot 3rd note, aaj aanand 1st note, and scarborough fair
create  b%maj8   C_   D_   e%_   F_   G_   A_   b%_   C'   write-key
: b%maj b%maj8 set-key ;
: kafi  b%maj ;

\ asavari  = raga with three flats on B, E, A       EQUIVALENT TO E%-MAJ
create     e%maj8 b%__ c_ d_ e%_ f_ g_ a%_ b%_ write-key  \ 1st key on b%  stairway-to-heaven
: e%maj    e%maj8 set-key ; 
create     asavari8  C_ D_ e%_ F_ G_ a%_ b%_ C' write-key
: asavari  asavari8 set-key ; 

\ bhairavi  = raga with four flats on B, E, A, D    EQUIVALENT TO A%-MAJ
create      a%maj8  C_   d%_   e%_   F_   G_   a%_   b%_   C'  write-key
: a%maj     a%maj8 set-key ;
: bhairavi  a%maj ;

\ kalyan  = raga with 1 sharp on F   WESTERN: G MAJOR
create   gmaj8   C_    C_    E_    f#_    G_    A_    B_    C'  write-key
: gmaj   gmaj8 set-key ;
: kalyan gmaj ; 

\ purvi   = raga with 1 sharp on F and two flats on D and A,    NO EQUIVALENT WESTERN MAJOR SCALE   ex: "back to life, back to reality"
            \  (close to A-MAJOR although the 3 decorated notes are also 3 sharps on F, C, G,  the scale also has C and G natural)
create purvi8   C_    d%_     E_    f#_    G_    a%_    B_    C' write-key
: purvi  purvi8 set-key ; 
   
\ todi  = it's a bhairavi raga (four flats on B, E, A, D ) with additionally 1 sharp on F  NO EQUIVALENT IN WESTERN SCALE
            \ (close to D%-MAJ  -- the f# is taken as G%, but also preserves the G natural )
create todi8   a%__   b%__   C_     d%_    e%_     f#_     G_     A%_  write-key  \ starting at A%
: todi  todi8 set-key ;
create todiC8   C_    d%_    e%_    f#_    G_    a%_    b%_    C'   write-key
: todiC  todiC8 set-key ;

\ bilawal = =raga with no flats or sharps, ex: abadoo (on C) and doh-a-deer  \ Western C-major 
create bilawal8  G__ A__ B__ C_ D_ E_ F_ G_  write-key  \ transposed to G  ex: abadoo starting on 4
: bilawal bilawal8  set-key ; 


\ ALGERIAN: https://www.pianoscales.org/algerian.html
\ Example algerian electric guitar: https://www.guitar9.com/column/why-does-scale-sound-exotic
\  2 1 2  1 1 1  3 1
create algerian8  ( D_ ) E_ F_ G_ A%_ A_ B%_ C#' D'  write-key
: algerian  algerian8 set-key ;

\ Arabic Music: http://www.maqamworld.com/en/index.php
\ Jins and Maqams
\ JINS: http://www.maqamworld.com/en/jins.php
create hijaz8  D_ E%_ F#_ G_   A_ B%_ C#' D' write-key  \ 1 3 1 2
: jins-hijaz hijaz8 set-key ;
create kurd8 D_ E%_ F_ G_   A_ B%_ C' D' write-key    \ 1 1.5 .5 2   \ 2nd note tuned high, third note tuned low
: jins-kurd  kurd8 set-key ;
create lami8 D_ E%_ F_ G_   A%_ B%_ C' D' write-key
: jins-lami lami8 set-key ;

create selisir8  C_ D%_ E%_ G_ A%_ C' D%' E%' write-key
: balinese selisir8  set-key ;

create chinese8  C_ E_ F#_ G_ B_ C' E' F#' write-key
: chinese chinese8 set-key ;

create mongolian8  C_ D_ E_ G_ A_ C' D' E' write-key
: mongolian mongolian8 set-key ;

\  NON-WESTERN FREQUENCIES
\ Exotic scales with intervals shorter than a half-step: https://www.guitar9.com/column/why-does-scale-sound-exotic
\ Sruti (Indian)
\ Jins (Arabic)


\ CHINESE SCALES:  Nanyin performance: http://2centsmusic.blogspot.com/2011/05/nanyin.html
\ Theory: https://web.archive.org/web/20131009041743/http://2centsmusic.blogspot.se/2011/04/tonality-and-scales-in-chinese-music.html
\ Yu scale on Huang Zhong base freq ( Why Don't you come back... and Fire on the Mountain)
create yu8approx A_ C' D' E' G' A' C'' D'' write-key
: yu-approx  yu8approx set-key ;
create yu8Gapprox G_ A_ C' D' E' G' A' C'' write-key  \ transposed 1 note lower for Fire on the mountain, starting on key#2
: yuG-approx  yu8Gapprox set-key ;

create yu8  yu-root write-freq-scale
: yu  yu8 set-key ;
create shang8  shang-root write-freq-scale
\ shang from note 4 is like yu but higher
: shang  shang8 set-key ;
\ Jue from note 3 is like yu but higher
create jue8  jue-root write-freq-scale
: jue  jue8  set-key ; 
\ Zhi from note 2 is like yu
create zhi8  zhi-root write-freq-scale
: zhi  zhi8  set-key ;


\ ================ APPLICATION - string instrument tunings
\ Instrument tunings: https://onlinetonegenerator.com/tuning.html
create guitar8  E__ A__ D_ G_ B_ E' A' D'' write-key 
: guitar  guitar8 set-key ;
create bass8    E___ A___ D__ G__ E__ A__ D_ G_ write-key  \ first 4 are bass strings, 1 octave lower than the bass of a 6-string guitar
: bass    bass8   set-key ;
create ukelele8 G_ C_ E_ A_ G' C' E' A' write-key     \ top G(4) C(3) E(2) A(1) highest - soprano/concert ukelele, next 4 are given an octave higher
: ukelele  ukelele8 set-key ;
create violin8  G__ D_ A_ E' G_ D' A' E'' write-key   \ first note is lowest string of viola, then violin 4 notes
: violin  violin8 set-key ; 
create cello8  C___ G___ D__ A__ C__ G__ D_ A_ write-key \ first 4 are cello, next 4 are viola
: cello-viola  cello8 set-key ;

\ ======== SETUP

: init-keys  \ this is now in Setup code of Piano.avr
   10 13 setreg  \ set min duration (ms) of note
   5 14 setreg   \ set settle time following note
;

\ DESIGN: not needed since Piano.avr starts up with cstd scale
: init-scale  cstd8 0 transpose8 write-scale ;

: start-speaker ( pin -- )  \ programmable digital synthesizer
    dup out
    setspeaker
	cstd play-scale  ;    \ nice chord: b.. d. g. 

\  : start-piano4x4  ( pin -- ) \ programmable digital synthesizer PCB with lights board 
\       start-speaker
\       out4 use-port-out  ;



: help-scales
  cr ." Western styles: equitemper  cstd  cmaj"
  cr ." Chinese styles (5): yu yu-1(*) shang gong jue zhi"
  cr ." Indian styles (8): bhairav, asavari, bhairavi, kafi, khamaj, kalyan, purvi, todi"
;


: help-misc
    cr ."           440 tone      - sends 440 Hz (A) note to speaker"
    cr ."    c. d. e. c' c'' c.. c... f'# a.% -  sends tone to pin 13 (speaker)  c. is middle, c.. is octave lower, c' is octave higher, # sharp, % flat"
    cr ."    1 .. 2 ..  - convert numbered simple piano to notes"
    cr
    cr ."  playlist   - runs through wings, harp, ode, memory, daffodils, morning, stairway, freedom, rainbow, reason, scarborough "
    cr ."  compositions: Assad: yearning snake  Jasmine: tog (together) "
    cr ."  => SOUND / MUSIC (tone)   include 3iSounds.fs"  
    cr ."       13 start-speaker  - sets pin 13 to speaker" 
    cr ."       record         - toggles recording mode for the keys (turns digital piano serial transmit on/off)"
    cr ."       save-bar       - saves 5 notes"
    cr ."       play-notes     - plays all saved notes"
    cr ."       delete-note    - delete last saved note"
;

: help-sound
    cr
    cr ." HELP-SOUND"
    cr ." ----------"
    cr
    cr ." Get started:"
    cr ."     plug in digital piano"
    cr ."     use RealTerm to open the connection to the port"
    cr ." In GForth:"
    cr ."     include 3iForth.fs   include 3iPort4x4.fs   include 3iSound4x4.fs"
    cr ."     1 open-com  13 start-speaker   or   13  start-piano4x4"
    cr ." Several ways to play:"
    cr ."    1.  440 tone  - play  frequency 440Hz A"
    cr ."    2.  45 note - play 45th note (440Hz A) from zero point, organized in octaves of 12 notes each"
    cr ."    3.  c_ note   c__ note   c' note   c'' note    c#__ note  b%__  note  - plays note, any octave, sharp or flat"
    cr ."    4.  1 ,,  2 ,,  - plays the diatonic (white key) notes from current scale (8 preferred keys)"
    cr ." Programming scales/key-signatures:"
    cr ."    5.  ><<   >>>  - transpose keyboard by an octave up or down "
    cr ."    6.  cstd  1flat 2flats 3flats 4flats 5flats 1sharp 2sharps 3sharps 4sharps 5sharps - Western scales"
    cr ."    7.  bhairav  yu  
    cr ."    7.  play-scale   "
    cr ."    9.  cstd  1 2 3  1 3 1 3  7 play-bar - plays n notes from stack"
    cr
    cr ."   10.  cstd8 show-scale"
    cr ."   11.  create name8 ... write-key     name8 set-key"
    cr ."   12.  1 buttonsrecord  (turns on serial notification of button press, see with rx)"
    cr ." More help:  help-scales  help-misc  help-record   help-playlist"
;


\ ****************************
help-sound


\ =======================================
\ Version History
\ V2 - 2020-10-03---Sat---22:00 - Expanded to include the lights capability with Piano-4x4
\ V3 - 2021-05-16---Sun---17:14 - perfected and simplified the multi-octave notation with the piano-lights board 32-keys, playlist works with play that also teaches with lights
\ V4 - 2021-08-13---Fri---04:37 - cleaning up scales/key signatures design to make it much simpler to create with Indian raga scales --- RESULT: an 8-key concept with 4 buttons
\ V5 - 2021-08-17---Tue---21:23 - encoded Chinese scales directly into frequency tables


\ IDEA:  string instrument tuner..
\ IDEA:  song prompter for guitar...
\ http://mathscitech.org/music.html
